self.__BUILD_MANIFEST=function(s,e){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":[s,"static/chunks/818-50255d3da5031c79.js",e,"static/css/91ca400e43b3481b.css","static/chunks/pages/index-235bdf9d19b1f814.js"],"/_error":["static/chunks/pages/_error-7a92967bea80186d.js"],"/benchmarks":[s,e,"static/css/0096c84fa71d6718.css","static/chunks/pages/benchmarks-5e7a4d2f1223ad77.js"],sortedPages:["/","/_app","/_error","/benchmarks"]}}("static/chunks/808-ba05144b7460a1cd.js","static/chunks/472-9ebf9024bc039b81.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();